(* $Id: godi_profile_ops.ml 1096 2012-10-15 14:55:17Z gerd $ *)

open Godi_types
open Godi_exec
open Printf

let list() =
  let d = 
    Godi_file.addname
      (Godi_query.buildbase_filename())
      "profiles" in
  Godi_profile.list_profiles d


let commented_list() =
  let d = 
    Godi_file.addname
      (Godi_query.buildbase_filename())
      "profiles" in
  let names = Godi_profile.list_profiles d in
  List.map
    (fun name ->
       let comment =
         try
           let p = 
             Godi_profile.read_profile_ini
               (Godi_file.addname (Godi_file.addname d name) "profile.ini") in
           p.Godi_profile.prof_comment
         with
           | error ->
               "ERROR: Profile unparsable" in
       (name, comment)
    )
    names



let enable name =
  let d = 
    Godi_file.addname
      (Godi_query.buildbase_filename())
      "profiles" in
  let p =
    Godi_profile.read_profile_ini
      (Godi_file.addname
	 (Godi_file.addname d name)
	 "profile.ini") in
  (* 1. modify godi.conf *)
  ignore(Godi_conffile.modify_godi_conf
	   ~pref_user_error:true
	   ~opts:["GODI_PROFILE", name]
	   []
	);
  (* 2. write profile.conf *)
  Godi_profile.write_profile_conf
    (Godi_file.filename (Godi_query.profile_conf())) p;
  (* 3. update profile patches *)
  Godi_profile.write_profile_patches
    (Godi_file.addname
       (Godi_query.buildbase_filename())
       "profile-patches")
    p


let current() =
  let profile_ini =
    Godi_file.addname (Godi_query.profile_filename()) "profile.ini" in
  Godi_profile.read_profile_ini profile_ini


let update con =
  let p = current() in
  ignore(Godi_update.update_buildfiles con p)


let extract ?(constraints=[]) con =
  let p = current() in
  (* We add the ad-hoc constraints to all sources *)
  let prof_sources =
    List.map
      (fun src ->
         { src with
             Godi_profile.prof_include = (src.Godi_profile.prof_include @ 
                                            constraints)
         }
      )
      p.Godi_profile.prof_sources in
  let p1 = 
    { p with Godi_profile.prof_sources = prof_sources } in
  Godi_update.extract_buildfiles con p1


let wish con =
  let wishes = ref (Godi_wishes.read_wishes ~pref_user_error:true ()) in
  let modified = ref false in
  let src = Godi_query.scan_source_packages con in
  let inst = Godi_query.get_installed_packages con in
  let inst_wish =
    List.filter
      (fun pkg ->
	 let name = pkg_name pkg in
	 let fullname = pkg_fullname pkg in
	 let src_fprint = pkg_src_fingerprint pkg in
	 let ex = 
	   List.exists
	     (fun spkg -> 
		(* The src_fingerprint changes when the build.tgz file
		   is modified. This is the strictest criterion here.
		   We do this only if we have this information.
		 *)
		pkg_name spkg = name && 
		 (pkg_fullname spkg <> fullname ||
		    match src_fprint, pkg_src_fingerprint spkg with
		      | Some fp1, Some fp2 -> fp1 <> fp2
		      | _ -> false
		 )
	     )
	     src in
	 ex
      )
      inst in
  List.iter
    (fun pkg ->
       let wish = `Build (pkg_name pkg) in
       if not (List.mem wish !wishes) then (
	 wishes := !wishes @ [ wish ];
	 modified := true
       )
    )
    inst_wish;
  if !modified then
    Godi_wishes.write_wishes ~pref_user_error:true !wishes


let copy oldname newname =
  let d = 
    Godi_file.addname
      (Godi_query.buildbase_filename())
      "profiles" in
  let p =
    Godi_profile.read_profile_ini
      (Godi_file.addname
	 (Godi_file.addname d oldname)
	 "profile.ini") in
  let p' =
    { p with 
        Godi_profile.prof_name = newname; 
        prof_origin = None;
        prof_comment = "Copy of: " ^ p.Godi_profile.prof_comment;
    } in
  let oldpath = Godi_file.addname d oldname in
  let newpath = Godi_file.addname d newname in
  let l = Godi_file.read_dir_hierarchy oldpath in
  Godi_file.remove ~recurse:true newpath;
  List.iter
    (fun n ->
       let p = Godi_file.addname newpath n in 
       let d =
	 match Godi_file.dirname p with
	   | Some d -> d
	   | None -> assert false in
       Godi_file.create_dir ~recurse:true d;
       Godi_file.copy (Godi_file.addname oldpath n) p
    )
    l;
  Godi_profile.write_profile_ini
    (Godi_file.addname
       (Godi_file.addname d newname)
       "profile.ini")
    p'


let delete name =
  if name = Godi_query.profile_name() then
    raise(User_error
            "Godi_profile_ops.delete: cannot delete an enabled profile");
  let d = 
    Godi_file.addname
      (Godi_query.buildbase_filename())
      "profiles" in
  let path = Godi_file.addname d name in
  Godi_file.remove ~recurse:true path


let save_patches con =
  let d_patches =
    Godi_file.addname
      (Godi_query.buildbase_filename())
      "profile-patches" in
  let d = 
    Godi_file.addname
      (Godi_query.buildbase_filename())
      "profiles" in
  let p_name = Godi_query.profile_name() in
  let d_profile =
    Godi_file.addname
      (Godi_file.addname d p_name)
      "profile.ini" in
  let p = Godi_profile.read_profile_ini d_profile in
  if p.Godi_profile.prof_origin <> None then (
    printmsg con "Warning: The current profile is connected with a remote\n";
    printmsg con "site, and will be overwritten the next time packages\n";
    printmsg con "are updated! Your patches are then lost.\n"
  );
  let p' = Godi_profile.update_profile_patches d_patches p in
  Godi_profile.write_profile_ini d_profile p'


let wt_extract r =
  match r with
    | `Ok x -> x
    | `Timeout -> raise(User_error "Timeout during remote operation")
    | `Error msg -> raise(User_error msg)


let set_origin url name =
  let d = 
    Godi_file.addname
      (Godi_query.buildbase_filename())
      "profiles" in
  let path =
    Godi_file.addname
      (Godi_file.addname d name)
      "profile.ini" in
  let p1 = Godi_profile.read_profile_ini path in
  let p2 = { p1 with Godi_profile.prof_origin = Some url } in
  Godi_profile.write_profile_ini path p2

let unpack wt url name d =
  let tdir = Godi_file.allocate_tmpdir() in
  let files = wt # get tdir [ "/" ^ name ^ ".profile.tgz" ] in
  let file1 = wt_extract (List.hd files) in
  let file2 =
    match Godi_file.dirname file1 with
      | Some p2 ->
          let file2 = Godi_file.addname p2 (name ^ ".profile.tgz") in
          Godi_file.rename file1 file2;
          file2 
      | None ->
          assert false in
  Godi_profile.unpack_profile file2 d;
  Godi_file.release_tmpdir tdir;
  set_origin url name


let rget wtfac url name =
  (* Check for a local profile with [name] *)
  let d = 
    Godi_file.addname
      (Godi_query.buildbase_filename())
      "profiles" in
  let path =
    Godi_file.addname
      (Godi_file.addname d name)
      "profile.ini" in
  if Godi_file.file_exists path then (
    let p = Godi_profile.read_profile_ini path in
    let ok =
      match p.Godi_profile.prof_origin with
        | Some org -> org = url
        | None -> false in
    if not ok then
      raise(User_error
              "get: A local profile with the same name already exists");
  );
  (* Download and unpack *)
  let wt = wtfac # open_location url in
  unpack wt url name d

let rupdate wtfac name =
  (* Check for a local profile with [name] *)
  let d = 
    Godi_file.addname
      (Godi_query.buildbase_filename())
      "profiles" in
  let path =
    Godi_file.addname
      (Godi_file.addname d name)
      "profile.ini" in
  if not (Godi_file.file_exists path) then
    raise(User_error "update: there is no local profile with this name");
  let url = (
    let p = Godi_profile.read_profile_ini path in
    match p.Godi_profile.prof_origin with
      | Some org -> org
      | None ->
          raise(User_error
                  "update: this local profile does not have a remote URL")
  ) in
  (* Download and unpack *)
  let wt = wtfac # open_location url in
  unpack wt url name d


let rput wtfac url name =
  (* Check for a local profile with [name] *)
  let d = 
    Godi_file.addname
      (Godi_query.buildbase_filename())
      "profiles" in
  let path1 = Godi_file.addname d name in
  let path2 = Godi_file.addname path1 "profile.ini" in
  if not (Godi_file.file_exists path2) then
    raise(User_error "put: there is no local profile with this name");
  (* Pack *)
  let tdir = Godi_file.allocate_tmpdir() in
  let file = Godi_profile.pack_profile path1 tdir in
  (* Upload *)
  let wt = wtfac # open_location url in
  let () = wt_extract(wt # put file ("/" ^ name ^ ".profile.tgz")) in
  Godi_file.release_tmpdir tdir
  

let prof_re = Str.regexp "\\(.*\\)\\.profile\\.tgz$"


let rlist wtfac url =
  let wt = wtfac # open_location url in
  let files = wt_extract (wt # list "/") in
  let files1 =
    List.filter
      (fun n ->
         Str.string_match prof_re n 0
      )
      files in
  List.map
    (fun n ->
       ignore(Str.string_match prof_re n 0);
       Str.matched_group 1 n
    )
    files1


let rupdate_current con =
  let p = current() in
  match p.Godi_profile.prof_origin with
    | None ->
        printmsg con "Using a local-only profile (no profile update)\n"
    | Some url ->
        printmsg con
          (sprintf 
             "Updating profile from %s\n" url);
        let wtool = Wtool.wtool_factory() in
        wtool # configure_store_passwords (Godi_query.store_passwords());
        ( match Godi_query.user() with
            | None -> ()
            | Some u -> wtool # configure_user u
        );
        rupdate wtool p.Godi_profile.prof_name


