(* $Id: godi_query.ml 1118 2012-12-04 16:45:36Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_exec
open Godi_versions
open Godi_plist
open Printf


module StringSet = Set.Make(String) ;;
module StringMap = Map.Make(String) ;;


let tws_re = Str.regexp "[ \t\r\n]*$";;

let strip_trailing_ws s =
  Str.global_replace tws_re "" s
;;


let lazy_etc_dir =
  lazy (
    try Sys.getenv "GODI_CONFDIR"
    with
      | Not_found ->
	  let b = Buffer.create 100 in
	  exec_to_user_error
	    ~pref_user_error:true
	    (call_cmd ~stdout:(`Buffer b)) 
	    (shell ". godi_confdir");
	  strip_trailing_ws(Buffer.contents b)
  )
;;


let lazy_godi_conf =
  lazy (
    let etc_dir = Lazy.force lazy_etc_dir in
    etc_dir ^ "/godi.conf"
  )
;;


let lazy_profile_conf =
  lazy (
    let etc_dir = Lazy.force lazy_etc_dir in
    etc_dir ^ "/profile.conf"
  )
;;


let pcache = Godi_make.create_parser_cache();;


let from_godi_conf varname =
  let godi_conf = Lazy.force lazy_godi_conf in
  let ctx = 
    Godi_make.create_context
      ~make_cmd:"godi_make"
      [] in
  Godi_make.fast_parse_eval 
    ~pref_user_error:true
    ~sys_mk:false
    pcache
    godi_conf
    ctx;
  let loc = { Godi_make.loc_filename = "<Expansion of " ^ varname ^ ">";
	      Godi_make.loc_linenumber = 0 } in
  let v =
    try
      Godi_make.expand_var 
	~pref_user_error:true
	ctx
	loc
	varname
    with
      | Not_found ->
	  failwith("Variable " ^ varname ^ " not set in godi.conf")
  in
  strip_trailing_ws v
;;


let lazy_localbase =
  lazy (from_godi_conf "LOCALBASE")
;;


let lazy_pkgdbdir =
  lazy (from_godi_conf "GODI_DBDIR")
;;


let godi_conf() = Lazy.force lazy_godi_conf
let profile_conf() = Lazy.force lazy_profile_conf
let localbase() = Lazy.force lazy_localbase
let localbase_filename() = Godi_file.filename (localbase())
let pkgdbdir() = Lazy.force lazy_pkgdbdir
let pkgdbdir_filename() = Godi_file.filename (pkgdbdir())
let buildbase() = (localbase()) ^ "/build"
let buildbase_filename() = Godi_file.filename (buildbase())

let lazy_user =
  lazy(
    try Some(from_godi_conf "GODI_USER")
    with _ -> None
  )

let user() =
  Lazy.force lazy_user



let lazy_store_passwords =
  lazy (
    try from_godi_conf "GODI_STORE_PASSWORDS" = "yes"
    with _ -> false
  )

let store_passwords() =
  Lazy.force lazy_store_passwords


let get_tools_makefile() =
  let cmd_name = Sys.argv.(0) in
  let base_cmd_name = Filename.basename cmd_name in
  if base_cmd_name = "godi_console_dev" then (
    let bin_dir = Filename.dirname cmd_name in
    let f = bin_dir ^ "/../../mk/build/mk/console.mk" in
    if f.[0] = '/' then
      f
    else
      Filename.concat (Sys.getcwd()) f
  ) 
  else (
    match Godi_boot.stage() with
      | `Stage0 ->
	  failwith "Godi_query.get_tools_makefile: not possible in stage 0"
      | _ ->
	  let bbase = buildbase() in
	  bbase ^ "/mk/console.mk"
  )
;;


let from_makefile file varname =
  let etc_dir = Lazy.force lazy_etc_dir in
  let env =
    Array.concat [Unix.environment(); [| "GODI_CONFDIR=" ^  etc_dir |] ] in
  let ctx = Godi_make.create_context ~env (* ~curdir:dir ~objdir:dir *)
    ~make_cmd:"godi_make"
    ~override:[ "PACKAGE_SCANNING", "yes" ]
    [] 
    (* ?stderr:protocol *)
  in
  Godi_make.fast_parse_eval pcache file ctx;
  let loc = { Godi_make.loc_filename = "<Expansion of " ^ varname ^ ">";
	      Godi_make.loc_linenumber = 0 } in
  Godi_make.expand_var ctx loc varname
;;


let from_bsd_prefs_mk varname =
  match Godi_boot.stage() with
    | `Stage0 ->
	raise Not_found
    | _ ->
	let file = (buildbase()) ^ "/mk/bsd.prefs.mk" in
	from_makefile file varname
;;


let from_ocaml_mk varname =
  let file = (localbase()) ^ "/lib/godi/ocaml.mk" in
  if Godi_file.file_exists (Godi_file.filename file) then
    from_makefile file varname
  else
    raise Not_found
;;


let shlib_type() =
  try from_bsd_prefs_mk "SHLIB_TYPE" with Not_found -> "NA" ;;

let wget() =
  try from_bsd_prefs_mk "WGET" 
  with Not_found -> failwith "Cannot get WGET" ;;

let wget_options() =
  try from_bsd_prefs_mk "WGET_OPTIONS" 
  with Not_found -> ""  (* This is not an error *) ;;

let profile_name() =
  let n =
    try from_bsd_prefs_mk "GODI_PROFILE"
    with Not_found -> "" in
  if n="" then failwith "The variable GODI_PROFILE is unset!";
  n

let profile_filename() =
  let n = profile_name() in
  Godi_file.addname
    (Godi_file.addname (buildbase_filename()) "profiles")
    n

let ws_re = Str.regexp "[ \t]+"

let godi_binpkg_path() =
  try Str.split ws_re (from_bsd_prefs_mk "GODI_BINPKG")
  with Not_found -> []


let have_ocamlopt() = 
  try from_ocaml_mk "GODI_HAVE_OCAMLOPT" = "yes" with Not_found -> false ;;

let have_ocamloptp() = 
  try from_ocaml_mk "GODI_HAVE_PROF" = "yes" with Not_found -> false ;;

let have_ocamldll() = 
  try from_ocaml_mk "GODI_HAVE_DLL" = "yes" with Not_found -> false ;;


let read_text_file ?(pref_user_error=false) ?default filename =
  (* Return the contents of [filename] *)
  if Godi_file.file_exists filename then
    try
      Godi_file.read_file ~mode:`Text filename
    with
      | Sys_error msg when pref_user_error ->
	  user_error("System error: " ^ msg)
  else
    match default with
      | None -> 
	  if pref_user_error then
	    user_error("File not found: " ^ Godi_file.slashname filename)
	  else
	    raise(Sys_error("File not found: " ^ Godi_file.slashname filename))
      | Some d -> d
;;


let read_optional_text_file ?pref_user_error filename =
  if Godi_file.file_exists filename then
    Some(read_text_file ?pref_user_error filename)
  else
    None
;;

(***********************************************************************)
(* Source packages                                                     *)
(***********************************************************************)

let lines_re = Str.regexp "\n";;

let sloppy_lines_re = Str.regexp "[ \t]*\n[ \t]*";;

let words_re = Str.regexp "[ \t]+";;

let dep_re = Str.regexp "\\([^:]*\\):";;



let parse_dep_string s =
  if Str.string_match dep_re s 0 then (
    let cond_str = Str.matched_group 1 s in
    parse_dependency cond_str
  )
  else
    parse_dependency s
;;


let parse_dep_strings s =
  let words = Str.split words_re s in
  List.map parse_dep_string words
;;


let query_makefile ?pref_user_error ?godi_conf:gcarg ?(makefile="Makefile") 
                   ?(uncached=false) con dir variables =

(*
  let print_error msg =
    match protocol with
	None ->
	  prerr_string msg;
	  flush stderr
      | Some p ->
	  Godi_exec.collect p msg
  in
 *)

  let gc =
    match gcarg with
	None    -> godi_conf()
      | Some cf -> cf
  in

  let env =
    Array.concat [ Unix.environment(); 
		   [| "GODI_CONF=" ^  gc;
                      "GODI_CONFDIR=" ^ Filename.dirname gc;
                   |] 
		 ] in
  (* Latter variables override previous instances ! *)

  (* print_error ("Doing " ^ dir ^ "\n"); *)

  (* Notes:
   * - There is always GODI_CONFDIR in the environment
   * The result is that certain time-consuming actions need not to
   * be done when the Makefile is evaluated (running further external
   * programs)
   *)

  let ctx = Godi_make.create_context 
              ~env 
              ~curdir:(Godi_file.slashname dir)
              ~objdir:dir
	      ~make_cmd:"godi_make"
	      ~console:con
              ~override:[ "PACKAGE_SCANNING", "yes" ]
	      [] 
  in
  let pc =
    if uncached then
      Godi_make.create_parser_cache()
    else
      pcache in
  Godi_make.fast_parse_eval ?pref_user_error pc makefile ctx;
  List.map
    (fun varname ->
       let loc = { Godi_make.loc_filename = "<Expansion of " ^ varname ^ ">";
		   Godi_make.loc_linenumber = 0 } in
       Godi_make.expand_var ?pref_user_error ctx loc varname
    )
    variables
;;


let read_confopts dir =
  let fn = Godi_file.addname dir "CONFOPTS" in
  if Godi_file.file_exists fn then
    let c = read_text_file fn in
    List.filter (fun line -> line <> "") (Str.split sloppy_lines_re c)
  else
    []
;;


module VId = struct
  type t = var_id
  let compare = (compare : t -> t -> int)
end;;

module VIdMap = Map.Make(VId);;


class source_package_impl ?pref_user_error con (dir:Godi_file.filename) 
                          : source_package =
  let confopts = read_confopts dir in
  let values =
    query_makefile
      ?pref_user_error
      con
      dir (["PKGNAME"; "COMMENT"; "MAINTAINER"; "DEPENDS"; "BUILD_DEPENDS"; 
	    "CONFLICTS"; "MESSAGE_SRC"; "SELF_UPDATE_PKG";
	   ] @ confopts) in

  let (fullname,comment,maintainer,depends,build_depends,conflicts,
       message_src, self_upd_pkg, confopts_vals) =
    match values with
	fullname :: comment :: maintainer :: depends :: build_depends :: 
	  conflicts :: message_src :: self_upd_pkg :: confopts_vals ->
	    (fullname, comment, maintainer, depends, build_depends, conflicts, 
	     message_src, self_upd_pkg, confopts_vals)
      | _ ->
	  assert false
  in

  (* CHECK: Maybe we should catch errors in the following expressions: *)
  let (name, version, pkgrevision) = split_name_version fullname in
  let init_dependencies = parse_dep_strings depends in
  let init_build_dependencies = parse_dep_strings build_depends in
  let init_conflicts = parse_dep_strings conflicts in

object(self)
  val mutable user_status = `Keep
  val mutable user_options = None

  val dependencies = init_dependencies
  val build_dependencies = init_build_dependencies
  val conflicts = init_conflicts

  val mutable config_dialog_shown = false

  val var_id = new_var_id()
  val mutable variants = VIdMap.empty

  initializer
    variants <- VIdMap.add var_id self variants

  method name = name
  method version = version
  method pkgrevision = pkgrevision
  method comment = comment
  method godi_maintainer = maintainer
  method description =
    read_text_file ?pref_user_error ~default:"" 
      (Godi_file.addname dir "DESCR")

  method src_fingerprint =
    read_optional_text_file ?pref_user_error 
      (Godi_file.addname dir ".src_fingerprint")

  method dependencies = dependencies
  method build_dependencies = build_dependencies
  method conflicts = conflicts
  method var_id = var_id

  method self_update_pkg = 
    if self_upd_pkg = "" then
      None
    else
      Some(Godi_file.filename self_upd_pkg)

  method variant vid =
    try
      VIdMap.find vid variants
    with
	Not_found ->
	  failwith "class source_package: No such variant"

  method user_status = user_status
  method set_user_status s =
    match s with
	`Keep | `Build | `Build_dep _ ->
	  user_status <- s
      | _ ->
	  failwith "Bad user_status"

  method source_directory = dir

  method build_message =
    read_text_file ?pref_user_error ~default:"" 
      (Godi_file.addname dir "BUILDMSG")

  method install_message =
    read_text_file ?pref_user_error ~default:"" 
      (Godi_file.addname dir message_src)

  method build_options =
    confopts

  method values_of_build_options =
    List.combine confopts confopts_vals

  method values_of_user_build_options =
    match user_options with
	Some opts -> opts
      | None -> self#values_of_build_options

  method set_user_build_options opts =
    let names = List.map fst opts in
    if names <> self#build_options then
      invalid_arg "set_user_build_options";

    user_options <- Some opts

  method config_dialog_shown = config_dialog_shown

  method set_config_dialog_shown b = config_dialog_shown <- b

  method rescan_deps con godi_conf_file vid =
    let values =
      query_makefile
	?pref_user_error
	~godi_conf:godi_conf_file
	con
	dir [ "DEPENDS"; "BUILD_DEPENDS"] in

    let v_deps, v_bld_deps =
      match values with
	  [ v_deps; v_bld_deps ] -> (v_deps, v_bld_deps)
	| _ -> assert false in

    let new_pkg =
      {< dependencies = parse_dep_strings v_deps;
	 build_dependencies = parse_dep_strings v_bld_deps;
	 var_id = vid
      >} in

    variants <- VIdMap.add vid new_pkg variants;
    ()

end
;;


exception Next of string list
;;

let list_subdirs dirname =
  (* Return the list of all subdirectories of [dirname]; the returned names
   * include [dirname].
   *)
  List.filter
    Godi_file.file_is_dir
    (List.map
       (Godi_file.addname dirname)
       (Godi_file.read_dir dirname)
    )
;;


let really_scan_source_packages con =
  (* To get a list of all source packages, traverse the subdirectories
   * at LOCALBASE/build. All directories two levels below this location
   * that have a "Makefile" and a "DESCR" file are considered as
   * source package.
   * This is a lousy definition!
   *)
  let build = (localbase()) ^ "/build" in
  (* TODO: Handle errors from list_subdirs: *)
  let level1 =
    list_subdirs (Godi_file.filename build) in
  let level2 =
    List.flatten
      (List.map (fun dirname -> list_subdirs dirname) level1) in
  let source_dirs =
    List.filter
      (fun dirname ->
	 Godi_file.file_exists(Godi_file.addname dirname "Makefile") &&
	 Godi_file.file_exists(Godi_file.addname dirname "DESCR"))
      level2 in

  List.flatten
    (List.map
       (fun dir ->
	  try
	    let src_pkg = new source_package_impl 
			    ~pref_user_error:true con dir in
	    [ `Source_pkg src_pkg ]
	  with
	      User_error msg ->
		printmsg con
		  (sprintf
		     "Scan of directory %s: %s\n" 
		     (Godi_file.slashname dir) msg);
		[]
       )
       source_dirs)
;;


let source_packages = ref None;;


let scan_source_packages ?(rescan=false) con =
  if rescan then source_packages := None;
  match !source_packages with
      None ->
	let pkg_list = really_scan_source_packages con in
	source_packages := Some pkg_list;
	pkg_list
    | Some pkg_list ->
	pkg_list
;;


let forget_source_packages() =
  source_packages := None
;;


(***********************************************************************)
(* Installed packages                                                  *)
(***********************************************************************)

let name_re = Str.regexp "^@name[ \t]+\\([^ \t]+\\)[ \t]*$"
let pkgdep_re = Str.regexp "^@pkgdep[ \t]+\\([^ \t]+\\)[ \t]*$"
let blddep_re = Str.regexp "^@blddep[ \t]+\\([^ \t]+\\)[ \t]*$"


(* TODO: Currently we decompress the whole tgz. Use camlzip and avoid
   that
 *)

let from_tgz tgz =
  let components = Hashtbl.create 10 in
  ( try
      let ar = Pax.read (`File tgz) `Gzip `Tar in
      let cont = ref true in
      while !cont do
	match Pax.read_element ar with
	  | None -> cont := false
	  | Some el -> 
	      let m = Pax.get_metadata el in
	      let name = m#name in
	      if name <> "" && name.[0] = '+' then
		Hashtbl.replace 
		  components 
		  name
		  (Pax.extract_element_to_memory el)
	      else
		cont := false
      done;
      Pax.close_in ar;
    with _ -> ()
  );
  ( object
      method filename = tgz
      method get name =
	try Some(Hashtbl.find components name)
	with Not_found -> None
    end
  )
;;


let parse_CONTENTS ?(pref_user_error=false) ?no_files dir =
  let filename = Godi_file.addname dir "+CONTENTS" in
  try
    parse_inst_plist ?no_files filename
  with
      Sys_error msg when pref_user_error ->
	user_error("System error: " ^ msg)
    | Plist_error(file, line) when pref_user_error ->
	user_error("Bad line in file `" ^ file ^ "': " ^ line)
;;


let parse_CONTENTS_from_tgz ?(pref_user_error=false) ?no_files tgz_obj =
  (* This version is used by [binary_packege_impl] *)
  let plist_str = 
    match tgz_obj # get "+CONTENTS" with
	Some s -> s
      | None ->
	  user_error("This file is not a binary GODI package: " ^ 
		       Godi_file.slashname(tgz_obj # filename))
  in
  try
    parse_inst_plist_string ?no_files plist_str
  with
    | Plist_error(file, line) when pref_user_error ->
	user_error("Bad line in `+CONTENTS' in `" ^ 
		     Godi_file.slashname(tgz_obj#filename) ^ 
		     "': " ^ line)
;;


let query_package_name ?(pref_user_error=false) plist =
  try Godi_plist.extract_package_name plist
  with
    | Not_found when pref_user_error ->
	user_error("Missing @name directive in +CONTENTS")
;;


type inst_set_status =
    [ `Reload
    | `Reconcile of package list
    | `Up_to_date of package list
    ]
;;


let installed_packages = ref (`Reload : inst_set_status);;

let get_installed_packages_fn = ref (fun _ -> assert false) ;;
  (* "forward" declaration *)


class installed_package_impl ?pref_user_error (dir:Godi_file.filename) 
      : installed_package =
  let mtime = Godi_file.file_mtime dir in
  let plist = parse_CONTENTS ?pref_user_error (* ~no_files:true *) dir in
  let fullname = query_package_name ?pref_user_error plist in
  let dependencies = Godi_plist.extract_pkgdep plist in
  let version_deps =
    List.flatten
      (List.map
         (function
            | (name, Binary(_,fp)) -> [name,fp]
            | _ -> []
         )
         dependencies
      ) in
  let build_dependencies = Godi_plist.extract_blddep plist in
  let conflicts = Godi_plist.extract_pkgcfl plist in
  let src_fingerprint = Godi_plist.extract_src_fingerprint plist in
  let build_version_fingerprint = 
    Godi_plist.extract_build_version_fingerprint plist in

  let comment = 
    read_text_file ?pref_user_error (Godi_file.addname dir "+COMMENT") in
  let size =
    try Int64.of_string 
          (strip_trailing_ws 
	     (read_text_file (Godi_file.addname dir "+SIZE_PKG")))
    with
	Sys_error _ -> 0L
      | Failure _ ->   0L in

  let size_all =
    try Int64.of_string 
          (strip_trailing_ws
	     (read_text_file (Godi_file.addname dir "+SIZE_ALL")))
    with
	Sys_error _ -> 0L
      | Failure _ ->   0L in

  let preserve = Godi_file.file_exists (Godi_file.addname dir "+PRESERVE") in

  let [ req_script_path, req_script;
	inst_script_path, inst_script;
	deinst_script_path, deinst_script ] =
    List.map
      (fun b ->
	 let p = Godi_file.addname dir b in
	 if Godi_file.file_exists p then 
	   (Some p, Some (read_text_file ?pref_user_error p) )
	 else 
	   (None, None)
      )
      [ "+REQUIRE"; "+INSTALL"; "+DEINSTALL" ] in

  let (name, version, pkgrevision) = split_name_version fullname in

  let mapopt f =
    function
      | None -> None
      | Some x -> Some(f x) in

object(self:'self)
  val mutable user_status = `Keep

  method name = name
  method version = version
  method pkgrevision = pkgrevision
  method src_fingerprint = src_fingerprint
  method comment = comment
  method description = 
    read_text_file ?pref_user_error (Godi_file.addname dir "+DESC")
  method godi_maintainer = 
    read_text_file ?pref_user_error ~default:"" 
      (Godi_file.addname dir "+MAINTAINER")

  method db_directory = dir
  method preserve = preserve
  method inst_plist = parse_CONTENTS ?pref_user_error dir
  method size = size
  method size_all = size_all
  method requirements_script_path = mapopt Godi_file.slashname req_script_path
  method requirements_script = req_script
  method install_script_path = mapopt Godi_file.slashname inst_script_path
  method install_script = inst_script
  method deinstall_script_path = mapopt Godi_file.slashname deinst_script_path
  method deinstall_script = deinst_script

  method install_message =
    try 
      let f = Godi_plist.extract_display plist in
      read_text_file ?pref_user_error ~default:"" (Godi_file.addname dir f)
    with Not_found -> ""

  method build_info = 
    read_text_file ?pref_user_error ~default:"" 
      (Godi_file.addname dir "+BUILD_INFO")
  method build_version = 
    read_text_file ?pref_user_error ~default:"" 
      (Godi_file.addname dir "+BUILD_VERSION")

  method build_version_fingerprint =
    match build_version_fingerprint with
      | Some fp -> fp
      | None ->
	  (* compute it on the fly *)
          let s =
	    "MD5:" ^ Digest.to_hex(Digest.string self#build_version) in
          s


  method dependencies = dependencies
  method build_dependencies = build_dependencies
  method conflicts = conflicts
  method version_deps = version_deps

  method db_entry_last_modified = mtime

  method user_status = user_status
  method set_user_status s =
    match s with
	`Keep | `Remove | `Remove_dep _ ->
	  user_status <- s
      | _ ->
	  failwith "Bad user_status"

  method var_id = new_var_id()

  method variant (v:var_id) : 'self =
    failwith "class installed_package: No such variant"

  method update_successor_file ?(filter = fun _ -> true) con =
    (* Find out all packages that depend on this package: *)
    let all_installed = !get_installed_packages_fn con in
    let successors =
      List.filter
	(fun ipkg ->
	   (List.exists (fun (n,_) -> n = name) (pkg_dependencies ipkg)) &&
	   filter (pkg_name ipkg)
	)
	all_installed in
    Godi_file.write_file_lines 
      ~mode:`Binary
      (Godi_file.addname dir "+REQUIRED_BY")
      (List.map (fun succ -> pkg_fullname succ) successors);

end
;;


let update_installed_packages ?(verbose = false) con =
  (* To get a list of all source packages, traverse the subdirectories
   * at PKG_DBDIR. All subdirectories that have a "+CONTENTS" file
   * are considered as installed package.
   *)
  let dbdir = pkgdbdir() in
  (* TODO: Handle errors from list_subdirs: *)
  let level1 =
    list_subdirs (Godi_file.filename dbdir) in
  let inst_dirs =
    List.filter
      (fun dirname ->
	 Godi_file.file_exists(Godi_file.addname dirname "+CONTENTS"))
      level1 in

  let load_single_pkg dir =
    try
      let inst_pkg = new installed_package_impl 
		       ~pref_user_error:true dir in
      [ `Installed_pkg inst_pkg ]
    with
	User_error msg ->
	  printmsg con
	    (sprintf
	       "Scan of directory %s: %s\n" 
	       (Godi_file.slashname dir) msg);
	  []
  in

  let load dirs =
    List.flatten (List.map load_single_pkg dirs) in

  let update pl =
    (* Iterate over pl, and check whether the package exists and is
     * up-to-date. Returns a new list.
     *)
    List.flatten
      (List.map
	 (fun pkg ->
	    let inst_pkg = installed_pkg pkg in
	    if Godi_file.file_exists inst_pkg#db_directory then (
	      let mtime = Godi_file.file_mtime inst_pkg#db_directory in
	      if mtime <> inst_pkg#db_entry_last_modified then (
		load_single_pkg inst_pkg#db_directory  
                  (* Package was modified *)
	      )
	      else
		[ pkg ]   (* Package not modified *)
	    )
	    else
	      []   (* Package was deleted *)
	 )
	 pl
      )
  in

  let reconcile pl =
    (* First check the existing packages: *)
    let pl' = update pl in
    (* Now check whether there are new packages. Create a set of all known
     * directories, compare with inst_dirs
     *)
    let set =
      List.fold_left
	(fun acc pkg -> 
	   let dir = (installed_pkg pkg) # db_directory in
	   StringMap.add (Godi_file.slashname dir) dir acc)
	StringMap.empty
	pl' in
    let new_dirs =
      List.filter
	(fun dir -> 
	   not (StringMap.mem (Godi_file.slashname dir) set)) inst_dirs in
    pl' @ (load new_dirs)
  in

  match !installed_packages with
      `Reload ->
	(* printmsg con "Scanning system for installed packages...\n"; *)
	let pl = load inst_dirs in
	installed_packages := `Up_to_date pl
    | `Reconcile old ->
	(* printmsg con "Scanning installed packages for changes...\n"; *)
	let pl = reconcile old in
	installed_packages := `Up_to_date pl
    | `Up_to_date current ->
	()
;;


let get_installed_packages ?verbose con =
  update_installed_packages ?verbose con;
  match !installed_packages with
      `Up_to_date current ->
	current
    | _ ->
	assert false
;;


get_installed_packages_fn := get_installed_packages ;;


let get_this_installed_package ?verbose con name =
  let all_pkgs = get_installed_packages ?verbose con in
  List.find
    (fun pkg -> (pkg_name pkg) = name)
    all_pkgs
;;


let notify_about_change_of_installed_packages con =
  function
      `Reload ->
	installed_packages := `Reload
    | `Reconcile ->
	let pl = get_installed_packages con in
	installed_packages := `Reconcile pl
;;

(***********************************************************************)
(* Binary packages                                                     *)
(***********************************************************************)

let local_binary_package ?pref_user_error (tgz:Godi_file.filename) 
      : binary_package =
  let tgz = 
    match Godi_file.getcwd() with
      | None ->
	  user_error("Cannot determine path of working directory")
      | Some cwd ->
	  Godi_file.concat cwd tgz in
  let tgz_obj= from_tgz tgz in
  let tgz_plist =
    parse_CONTENTS_from_tgz ?pref_user_error (* ~no_files:true *) tgz_obj in
  let plist = 
    `Binpkg (Godi_file.basename tgz) ::
      (List.filter (function `Binpkg _ -> false | _ -> true) tgz_plist) in
  let fullname = query_package_name ?pref_user_error plist in
  let dependencies = Godi_plist.extract_pkgdep plist in
  let version_deps =
    List.flatten
      (List.map
         (function
            | (name, Binary(_,fp)) -> [name,fp]
            | _ -> []
         )
         dependencies
      ) in
  let build_dependencies = Godi_plist.extract_blddep plist in
  let conflicts = Godi_plist.extract_pkgcfl plist in
  let src_fingerprint = Godi_plist.extract_src_fingerprint plist in
  let build_version_fingerprint = 
    Godi_plist.extract_build_version_fingerprint plist in
  let descr_lz = lazy (
    match tgz_obj # get "+DESC" with
      | Some s -> s
      | None -> user_error("This file is not a binary GODI package: " ^ 
			     Godi_file.slashname tgz_obj#filename)
  ) in

  let comment = 
    match tgz_obj # get "+COMMENT" with
	Some c -> c
      | None   -> "" in

  let (name, version, pkgrevision) = split_name_version fullname in

object(self:'self)
  val mutable user_status = `Keep

  method name = name
  method version = version
  method pkgrevision = pkgrevision
  method comment = comment
  method description = Lazy.force descr_lz
  method godi_maintainer = "(Info n/a for non-source packages)"

  method binary_archive = tgz
  method require_binary_archive _ = ()
  method release_binary_archive() = ()

  method dependencies = dependencies
  method build_dependencies = build_dependencies
  method conflicts = conflicts
  method version_deps = version_deps

  method src_fingerprint = src_fingerprint

  method user_status = user_status
  method set_user_status s =
    match s with
	`Keep | `Add | `Add_dep _ ->
	  user_status <- s
      | _ ->
	  failwith "Bad user_status"

  method var_id = new_var_id()

  method variant (v:var_id) : 'self =
    failwith "class binary_package: No such variant"

  method inst_plist = plist

  method install_message = 
    try 
      let f = Godi_plist.extract_display plist in
      ( match tgz_obj # get f with
	  | None -> ""
	  | Some m -> m
      )
    with Not_found -> ""

  method build_version =
    match tgz_obj # get "+BUILD_VERSION" with
      | None -> "" 
      | Some s -> s

  method build_version_fingerprint =
    match build_version_fingerprint with
      | Some fp -> fp
      | None ->
	  (* compute it on the fly *)
	  "MD5:" ^ Digest.to_hex(Digest.string self#build_version)

  method build_info =
    match tgz_obj # get "+BUILD_INFO" with
      | None -> "" 
      | Some s -> s

  method requirements_script =
    tgz_obj # get "+REQUIRE"

  method install_script =
    tgz_obj # get "+INSTALL"

  method deinstall_script =
    tgz_obj # get "+DEINSTALL"

end
;;


let get_binary_package tgz =
  `Binary_pkg(local_binary_package ~pref_user_error:true tgz)
;;


let remote_binary_package ?pref_user_error plist url =
  let local_tgz = ref None in
  let local_pkg = ref None in
  let local_rel = ref None in
  let fullname = query_package_name ?pref_user_error plist in
  let dependencies = Godi_plist.extract_pkgdep plist in
  let version_deps =
    List.flatten
      (List.map
         (function
            | (name, Binary(_,fp)) -> [name,fp]
            | _ -> []
         )
         dependencies
      ) in
  let build_dependencies = Godi_plist.extract_blddep plist in
  let conflicts = Godi_plist.extract_pkgcfl plist in
  let src_fingerprint = Godi_plist.extract_src_fingerprint plist in
  let build_version_fingerprint = 
    Godi_plist.extract_build_version_fingerprint plist in
  let (name, version, pkgrevision) = split_name_version fullname in
  let comment = Godi_plist.extract_comment plist in
  let descr = Godi_plist.extract_description plist in

object(self:'self)
  val mutable user_status = `Keep

  method name = name
  method version = version
  method pkgrevision = pkgrevision
  method comment = comment
  method description = descr
  method godi_maintainer = "(Info n/a for non-source packages)"

  method binary_archive = 
    match !local_tgz with
      | None -> failwith "binary_archive: not locally available"
      | Some tgz -> tgz

  method require_binary_archive con =
    let (dir_url, base) = Godi_file_handler.split_name url in
    let handler = Godi_file_handler.handler_of_directory dir_url in
    let (tgz, release) = handler # get con base in
    let pkg = local_binary_package ?pref_user_error tgz in
    local_tgz := Some tgz;
    local_pkg := Some pkg;
    local_rel := release

  method release_binary_archive() =
    ( match !local_rel with
        | None -> ()
        | Some f -> f()
    );
    local_rel := None;
    local_tgz := None;
    (* can keep local_pkg *)

  method dependencies = dependencies
  method build_dependencies = build_dependencies
  method conflicts = conflicts
  method version_deps = version_deps

  method src_fingerprint = src_fingerprint

  method user_status = user_status
  method set_user_status s =
    match s with
	`Keep | `Add | `Add_dep _ ->
	  user_status <- s
      | _ ->
	  failwith "Bad user_status"

  method var_id = new_var_id()

  method variant (v:var_id) : 'self =
    failwith "class binary_package: No such variant"

  method inst_plist = 
    List.filter
      (function
         | `Binpkg _ | `Comment _ | `Descr _ -> false
         | _ -> true
      )
      plist

  method install_message = 
    match !local_pkg with
      | None -> failwith "install_message: package not locally available"
      | Some pkg -> pkg#install_message

  method build_version =
    match !local_pkg with
      | None -> failwith "build_version: package not locally available"
      | Some pkg -> pkg#build_version

  method build_version_fingerprint =
    match build_version_fingerprint with
      | Some fp -> fp
      | None ->
	  (* compute it on the fly *)
	  "MD5:" ^ Digest.to_hex(Digest.string self#build_version)

  method build_info =
    match !local_pkg with
      | None -> failwith "build_info: package not locally available"
      | Some pkg -> pkg#build_info

  method requirements_script =
    match !local_pkg with
      | None -> failwith "requirements_script: package not locally available"
      | Some pkg -> pkg#requirements_script

  method install_script =
    match !local_pkg with
      | None -> failwith "install_script: package not locally available"
      | Some pkg -> pkg#install_script

  method deinstall_script =
    match !local_pkg with
      | None -> failwith "deinstall_script: package not locally available"
      | Some pkg -> pkg#deinstall_script
end
;;
  

let get_remote_binary_package inst_plist url =
  `Binary_pkg(remote_binary_package ~pref_user_error:true inst_plist url)
;;


let split_manifest file =
  let parse cur =
    Godi_plist.parse_inst_plist_string
      (String.concat "\n" (List.rev cur)) in
  let rec split cur lines =
    match lines with
      | "" :: lines' ->
          if cur <> [] then
            parse cur :: split [] lines'
          else
            split [] lines'
      | line :: lines' ->
          split (line::cur) lines'
      | [] ->
          if cur <> [] then
            [ parse cur ]
          else
            []
  in
  let lines = Godi_file.read_file_lines file in
  split [] lines
;;


let scan_binary_packages con url =
  let handler = Godi_file_handler.handler_of_directory url in
  let pkgmanifest, rel = handler # get con "PKGMANIFEST"  in
  let pkgman_list = split_manifest pkgmanifest in
  ( match rel with
      | None -> ()
      | Some f -> f()
  );
  List.map
    (fun inst_plist ->
       let file = Godi_plist.extract_binpkg inst_plist in
       get_remote_binary_package
         inst_plist 
         (url ^ "/" ^ file)   (* FIXME: incorrect (URL encoding) *)
    )
    pkgman_list


let scan_all_binary_packages con =
  let p = godi_binpkg_path() in
  List.flatten
    (List.map
       (scan_binary_packages con)
       p
    )
;;  


let build_version_fingerprint pkg =
  match pkg with
    | `Binary_pkg p ->
        p#build_version_fingerprint
    | `Installed_pkg p ->
        p#build_version_fingerprint
    | `Dispersed_pkg p ->
        p#build_version_fingerprint
    | _ ->
        failwith "Godi_query.build_version_fingerprint: wrong pkg type"
;;


let version_deps pkg =
  match pkg with
    | `Binary_pkg p ->
        p#version_deps
    | `Installed_pkg p ->
        p#version_deps
    | `Dispersed_pkg p ->
        p#version_deps
    | _ ->
        failwith "Godi_query.version_deps: wrong pkg type"
;;


let binary_version_digest pkg =
  let bvf = build_version_fingerprint pkg in
  let vd = version_deps pkg in
  let dg =
    Digest.string
      (bvf ^ " " ^ 
         String.concat " "
           (List.map 
              (fun (n,fp) -> n ^ "/" ^ fp)
              vd)
      ) in
  String.sub
    (Digest.to_hex dg)
    0
    6


let binary_file_base pkg =
  match pkg with
    | `Binary_pkg _ 
    | `Installed_pkg _
    | `Dispersed_pkg _ ->
        pkg_fullname pkg ^ "@BUILD_" ^ binary_version_digest pkg
    | _ ->
        failwith "Godi_query.binary_file_base"

(***********************************************************************)


let matches_dep pkg (name,cond) =
  let p = any_pkg pkg in
  p#name = name && 
  version_matches_condition p#version cond &&
  ( match cond with
      | Binary(_,fp) ->
          ( match pkg with
              | `Binary_pkg _ | `Installed_pkg _ | `Dispersed_pkg _ ->
                  ( fp = build_version_fingerprint pkg || 
                      fp = "BUILD_" ^ binary_version_digest pkg
                  )
              | _ -> false
          )
      | _ -> true
  )
;;
